using System;
using System.Collections.Generic;
using System.Text;

namespace ima2wav
{

    /// <summary>
    /// Codec IMA-ADPCM (Encoder) 
    /// </summary>
    public class ImaAdpcmEncoder
    {
        // Internal state
        sbyte _index; // index for step table          __state
        int _pred;  // predicted signal (acumulated) __state

        sbyte[] _indexTable = { -1, -1, -1, -1, 2, 4, 6, 8, -1, -1, -1, -1, 2, 4, 6, 8 };
        short[] _deltaTable = { 7, 8, 9, 10, 11, 12, 13, 14, 16, 17, 19, 21, 23, 25, 28, 
            31, 34, 37, 41, 45, 50, 55, 60, 66, 73, 80, 88, 97, 107, 118, 130, 143, 157, 
            173, 190, 209, 230, 253, 279, 307, 337, 371, 408, 449, 494, 544, 598, 658, 
            724, 796, 876, 963, 1060, 1166, 1282, 1411, 1552, 1707, 1878, 2066, 2272, 2499, 
            2749, 3024, 3327, 3660, 4026, 4428, 4871, 5358, 5894, 6484, 7132, 7845, 8630, 
            9493, 10442, 11487, 12635, 13899, 15289, 16818, 18500, 20350, 22385, 24623, 
            27086, 29794, 32767};


        public ImaAdpcmEncoder()
        {
            // Set initial state
            Reset();
        }


        /// <summary>
        /// IMA-ADPCM Encoder
        /// </summary>
        public byte[] Encode(short[] ibuf)
        {
            // ibuf[i] input pcm sample
            //
            int   e;    // residue, to be quantized
            int   er;   // residue, reconstructed
            byte  eq;   // residue, quantized (only 4 LSB are used)            
            short step; // quantizer step
            byte  mask;

            byte[] obuf = new byte[ibuf.Length];

            for (int i = 0; i < ibuf.Length; i++)
            {
                // compute     the residue
                e = ibuf[i] - _pred;

                // quantize    the residue
                // reconstruct the residue
                if (e >= 0)
                    eq = 0;
                else
                {
                    eq = 8; // set sign bit
                    e = -e;
                }


                mask = 4;
                step = _deltaTable[_index];
                for (int j = 0; j < 3; j++)
                {
                    if (e >= step)
                    {
                        eq |= mask;
                        e -= step;
                    }

                    step >>= 1;
                    mask >>= 1;
                }

                obuf[i] = eq;

                er = 0;
                step = _deltaTable[_index];
                if ((eq & 4) != 0) er += step;
                if ((eq & 2) != 0) er += step >> 1;
                if ((eq & 1) != 0) er += step >> 2;

                er += (step >> 3);
                if ((eq & 8) != 0)
                    er = -er;

                // accumulate the residue (predicted sample)
                _pred += er;
                if (_pred > +32767)
                    _pred = +32767;
                else if (_pred < -32768)
                    _pred = -32768;

                // update the internal state (_index variable)
                _index += _indexTable[eq];
                if (_index < 00) _index = 00;
                if (_index > 88) _index = 88;
            }

            return obuf;
        }



        public byte[] Encode2(short[] ibuf)
        {
            int   e;     // residue, to be quantized
            int   er;    // residue, reconstructed
            byte  eq;    // residue, quantized (only 4 LSB are used)            
            short step;  // quantizer step
            

            byte[] obuf = new byte[ibuf.Length];

            for (int i = 0; i < ibuf.Length; i++)
            {
                // compute     the residue
                e    = ibuf[i] - _pred;
                step = _deltaTable[_index];

                Quantizer(e, step, out eq, out er);
                obuf[i] = eq;

                // accumulate the residue (predicted sample)
                _pred += er;
                if (_pred > +32767)
                    _pred = +32767;
                else if (_pred < -32768)
                    _pred = -32768;

                // update the internal state (_index variable)
                _index += _indexTable[eq];
                if (_index < 00) _index = 00;
                if (_index > 88) _index = 88;
            }

            return obuf;
        }



        public void Quantizer(int e, int step, out byte eq, out int er)
        {
            byte  mask;
            short temp;

            // compute     the residue

            // quantize    the residue
            // reconstruct the residue
            if (e >= 0)
                eq = 0;
            else
            {
                eq = 8; // set sign bit
                e = -e;
            }

            mask = 4;
            temp = (short) step;

            for (int j = 0; j < 3; j++)
            {
                if (e >= temp)
                {
                    eq |= mask;
                    e -= temp;
                }

                temp >>= 1;
                mask >>= 1;
            }

            //put eq;

            er = 0;
            
            if ((eq & 4) != 0) er += step;
            if ((eq & 2) != 0) er += step >> 1;
            if ((eq & 1) != 0) er += step >> 2;

            er += (step >> 3);
            if ((eq & 8) != 0)
                er = -er;

            //put er;
        }


        /// <summary>
        /// Reset internal state to zero (predicted sample acumulator and index register)
        /// </summary>
        public void Reset()
        {
            _index = 0;
            _pred = 0;
        }
    }
}
