///////////////////////////////////////////////////////////////////////////////////////////////////
// sndutil
// Utility to transform audio files, like sox but much more simple
//
// converts between wav files and raw files
// change sampling frequency
// normalize wav files
// etc
// much more comming
//
// 2008 Javier Valcarce-Garcia, <javier.valcarce@gmail.com>
///////////////////////////////////////////////////////////////////////////////////////////////////
using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.IO;


namespace wavutil
{
    static class Program
    {
        static FileStream istrm;
        static FileStream ostrm;

        static void Main(string[] args)
        {
            // Process command line arguments. This is a preliminar process that only checks what
            // options has been specified at command line and validates the data.
            // (Further processing may be required in main function)
            Args.CmdLineProc(args);

            /**/ if (Args.Flag_Help)
            {
                HelpMessage();
                return;
            }
            else if (Args.ErrorFlag)
            {
                Console.WriteLine("Error: " + Args.ErrorString);
                Console.WriteLine("Type -h for help");
                return;
            }


            // Main Options Dispatcher
            /**/ if (Args.Flag_Raw2Wav ) Opt_RawToWav();
            else if (Args.Flag_Wav2Raw ) Opt_WavToRaw();
            else if (Args.Flag_Resample) Opt_Resample();

        }



        //
        // Prints help message
        //
        static void HelpMessage()
        {

            // Display information about the EXE assembly.
        
            Console.WriteLine(Application.ProductName + " " + Application.ProductVersion + 
                " - WAV file utilities\n" + 
                "Copyright (C) 2008 Javier Valcarce-Garcia\n" +
                    "Usage: sndutil.exe [options] <input_file> <output_file>\n\n" +
                    "Options:\n" +
                    "  -h      Help\n" +
                    "  -s      Resample with a factor P/Q (Fs' = Fs * P/Q)\n" +
                    "  -w      Wav to raw file conversion\n" +
                    "  -r      Raw to wav file conversion\n" +
                    "  --fs=<Int32>  Sampling Frequency\n" +
                    "  --fmt=<X>     Raw file format (see below)\n" +
                    "  --P=<byte>    Interpolation factor\n" +
                    "  --Q=<byte>    Decimation factor\n\n" +
                    "Formats for raw file input/output:\n U08, S08, U16, S16, U24, S24, F32, F64");
        }


        /// <summary>
        /// Converts Wav file --> Raw file
        /// </summary>
        static void Opt_WavToRaw()
        {
            if (!Args.Flag_Fmt)
            {
                Console.WriteLine("--fmt parameter is mandatory");
                return;
            }

            // Create the streams and the associated readers and writers
            istrm = new FileStream(Args.Data_IFile, FileMode.Open);
            ostrm = new FileStream(Args.Data_OFile, FileMode.Create);

            WavFileReader wavrd = new WavFileReader(istrm);
            RawFileWriter rawwr = new RawFileWriter(ostrm, Args.Data_Fmt);

            Console.WriteLine("Wav --> Raw (fmt={0} Fs={1}Hz)",
                Args.Data_Fmt.ToString(), wavrd.Fs);

            if (!wavrd.Valid)
            {
                Console.WriteLine("Bad file format: {0}", istrm.Name);
                return;
            }

            int xfer;
            double[] buf = new double[4096];


            while (true)
            {
                xfer = wavrd.XferData(buf, 0, buf.Length);
                if (xfer == 0)
                    break;
                rawwr.XferData(buf, 0, xfer);
            }

            wavrd.Close();
            rawwr.Close();
        }


        /// <summary>
        /// Converts Raw file --> Wav file
        /// </summary>
        static void Opt_RawToWav()
        {
            if (!Args.Flag_Fmt || !Args.Flag_Fs)
            {
                Console.WriteLine("Error: --fmt and --fs parameters are mandatory");
                return;
            }

            // Create the streams and the associated readers and writers
            istrm = new FileStream(Args.Data_IFile, FileMode.Open);
            ostrm = new FileStream(Args.Data_OFile, FileMode.Create);

            RawFileReader rawrd = new RawFileReader(istrm, Args.Data_Fmt);
            WavFileWriter wavwr = new WavFileWriter(ostrm, WavFormat.FLT32,
                Args.Data_Fs);

            Console.WriteLine("Raw (fmt={0} Fs={1}Hz) --> WAV",
                Args.Flag_Fmt.ToString(), Args.Flag_Fs);

            int xfer;
            double[] buf = new double[4096];

            while (true)
            {
                xfer = rawrd.XferData(buf, 0, buf.Length);
                if (xfer == 0)
                    break;
                wavwr.XferData(buf, 0, xfer);
            }

            wavwr.Close();
            rawrd.Close();
        }


        /// <summary>
        /// Converts Wav file --> Raw file
        /// </summary>
        static void Opt_Resample()
        {
            if (!Args.Flag_P || !Args.Flag_Q)
            {
                Console.WriteLine("Error: --P and --Q parameters are mandatory");
                return;
            }


            // Create the streams and the associated readers and writers
            istrm = new FileStream(Args.Data_IFile, FileMode.Open);
            ostrm = new FileStream(Args.Data_OFile, FileMode.Create);

            WavFileReader wavrd = new WavFileReader(istrm);
            WavFormat wavfmt = wavrd.SampleFormat;

            int P = Args.Data_P;
            int Q = Args.Data_Q;

            int newfs = (int)Math.Round((double)wavrd.Fs * P / Q);
            WavFileWriter wavwr = new WavFileWriter(ostrm, wavfmt, newfs);

            int xlen;
            int order = 0064;
            int xsize = 1024;
            int ysize = (int)Math.Ceiling((double)xsize * P / Q);

            double[] xbuf = new double[xsize];
            double[] ybuf = new double[ysize];
            Resampler resampler = new Resampler(order, P, Q);
            Console.WriteLine("Resampling factor F = {0}/{1} = {2}", P, Q, (float)P / Q);
            Console.WriteLine("Resampling from {0}Hz to {1}Hz...", wavrd.Fs, newfs);
            int ylen;

            while (true)
            {
                xlen = wavrd.XferData(xbuf, 0, xbuf.Length);
                if (xlen == 0)
                    break;

                resampler.Filter(xbuf, xlen, ybuf, out ylen);
                wavwr.XferData(ybuf, 0, ylen);
            }

            wavwr.Close();
            wavrd.Close();
        }
    }
}
