using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Diagnostics;

namespace wavutil
{
    /// <summary>
    /// Changes sampling rate in a rational factor P/Q. Based on the classic method of 
    /// interpolation, reconstruction filter and a final decimation stage. 
    /// NOT VERY EFICCIENT
    /// </summary>
    public class Resampler
    {

        private double[] _h;    // reconstruction low-pass filter coefficients
        private double[] _z;    // delay line 
        private int _order;     // _z.Lenght-1
        private int _dptr;     // delay line pointer
        private int _skip;     // decimate counter
        private int _P;        // Interpolation Factor
        private int _Q;        // Decimation    Factor                
        


        #region Constructors
        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="order">Must be even</param>
        /// <param name="p"></param>
        /// <param name="q"></param>
        public Resampler(int order, int P, int Q)
        {
            if (order % 2 != 0)
                throw new Exception("order must be odd");

            // como el filtro es de fase lineal (la respuesta al impulso tiene un eje de simetra)
            // es necesario almacenar slo la mitad de los coeficientes ms la muestra central por
            // la que pasa el eje de simetra.
            _order = order;
            _z = new double[order     + 1];
            _h = new double[order / 2 + 1];
            _dptr = 0;
            _skip = 0;
            _P = P;
            _Q = Q;


            // cut frequency
            double wc = (_P > _Q) ? Math.PI / _P : Math.PI / _Q;

            // _P is necesary to normalize volume
            double h0 = _P*(wc / Math.PI);

            for (int i = 0; i < _h.Length; i++)
            {
                if (i == 0)
                    _h[_h.Length - 1 - i] = h0;
                else
                    _h[_h.Length - 1 - i] = h0 * Math.Sin(wc * i) / (wc * i);
            }
        }

        #endregion

        /// <summary>
        /// Filter a vector
        /// </summary>
        /// <param name="x">input sample vector</param>
        /// <param name="y">output sample vector</param>
        /// <param name="offset">offset in **both** vectors</param>
        /// <param name="count">sample count</param>
        public void Filter(double[] x, int xlen, double[] y, out int ylen)
        {
            double acum;
            double curr;

            int kx;
            int kv;
            int ky;
            int jf; // fordward index
            int jb; // backward index
            int i;

            Debug.Assert(y.Length == (int)Math.Ceiling((double)x.Length * _P / _Q));

            ky = 0;
            for (kx = 0; kx < xlen; kx++)
            {
                for (kv = 0; kv < _P; kv++)
                {
                    // add to delay line x[kx] followed by (_P-1) zeros
                    if (kv == 0)
                        curr = x[kx];
                    else
                        curr = 0.0;

                    // las muestras ms antiguas estn en posiciones de memoria decrecientes
                    _z[_dptr] = curr;
                    _dptr--; if (_dptr == -1) _dptr = _z.Length - 1;

                    // dptr points now to x(n-N+1)

                    // debido al diezmado por _Q, no es necesario computar todas y cada una
                    // de las muestras de entrada. Solo 1 de cada _Q
                    if (_skip < _Q - 1)
                    {
                        _skip++;
                        continue;
                    }
                    else
                    {
                        _skip = 0;
                    }

                    // --------------------------------------------------------------------
                    // Pasada de filtrado con h(n) simtrica

                    // fordward: x(n-1) x(n-2) ...
                    jf = _dptr;
                    jf++; if (jf == _z.Length) jf = 0;

                    // backward: x(n-N) x(n-N+1) ...
                    jb = _dptr;
                    acum = 0.0;

                    for (i = 0; i < _h.Length-1; i++)
                    {
                        // h simetric
                        acum += (_z[jf] + _z[jb]) * _h[i];

                        jf++; if (jf == _z.Length) jf = 0;
                        jb--; if (jb == -1       ) jb = _z.Length - 1;
                    }

                    jf++; if (jf == _z.Length) jf = 0;

                    // elemento central de h(n) por donde pasa el eje de simetra
                    acum += _z[jf] * _h[_h.Length - 1];
                    y[ky] = acum;
                    ky++;
                }
            }


            ylen = ky;
        }
    }
}